# -*- coding: utf-8 -*-

# Copyright 2008-2012 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys
import cgi
import textwrap
import re

DIRECTIVES = 'bcgituwz'

CASE_UPPER = 0
CASE_LOWER = 1
BASE_10 = 10
BASE_16 = 16

def _error(s):
    sys.stderr.write('\nERROR: %s\n' % s)
    sys.exit(1)

def parse_int(num_str):
    try:
        if num_str[0] in '$#&':
            return int(num_str[1:], 16)
        return int(num_str)
    except:
        return None

def get_address(operation, check_prefix=True):
    prefixes = '[ ,(+-]' if check_prefix else ''
    search = re.search('%s\$[0-9A-Fa-f]+' % prefixes, operation)
    if not search:
        search = re.search('%s[0-9]+' % prefixes, operation)
    if search:
        index = 1 if check_prefix else 0
        return search.group()[index:]

def set_bytes(snapshot, address, operation):
    directive = operation[:4].upper()
    if directive == 'DEFM':
        start = operation.index('"') + 1
        message = operation[start:operation.index('"', start)]
        snapshot[address:address + len(message)] = [ord(c) for c in message]
        return
    args = [parse_int(v) for v in operation[5:].split(',')]
    if directive == 'DEFS':
        span = args[0]
        value = args[1] if len(args) > 1 else 0
        snapshot[address:address + span] = [value] * span
    else:
        words = directive == 'DEFW'
        for arg in args:
            snapshot[address] = arg % 256
            address += 1
            if words:
                snapshot[address] = arg // 256
                address += 1

class SkoolParser:
    def __init__(self, skoolfile, case=None, base=None, asm_mode=0, verbose=False, warn=False, fix_mode=0, html=False, create_labels=False, debug=False):
        self.skoolfile = skoolfile

        # case:
        #   None = leave case alone
        #   0 = force upper case
        #   1 = force lower case
        # base:
        #   None = leave base alone
        #   10 = force decimal
        #   16 = force hexadecimal
        # asm_mode:
        #   0 = do not parse asm directives
        #   1 = parse asm directives in @isub mode
        #   2 = parse asm directives in @ssub mode
        #   3 = parse asm directives in @rsub mode
        # verbose:
        #   whether to display info about instruction substitutions
        # warn:
        #   whether to show warnings
        # fix_mode:
        #   0 = apply no fixes
        #   1 = honour @ofix directives
        #   2 = honour @ofix and @bfix directives
        #   3 = honour @ofix, @bfix and @rfix directives
        # html:
        #   whether to escape HTML characters
        # create_labels:
        #   whether to create default labels for unlabelled instructions
        self.mode = Mode(case, base, asm_mode, verbose, warn, fix_mode, html, create_labels)

        self.debug = debug

        self.snapshot = [0] * 65536  # 64K of Spectrum memory
        self.instructions = {}       # address -> [Instructions]
        self.entries = {}            # address -> Routine/Data
        self.memory_map = []         # Routines and Datas
        self.header = []
        self.stack = []
        self.comments = []
        self.ignores = []

        if skoolfile == '-':
            self._parse_skool(sys.stdin)
        else:
            f = open(skoolfile, 'r')
            self._parse_skool(f)
            f.close()

    def get_word(self, address):
        return self.snapshot[address] + 256 * self.snapshot[address + 1]

    def get_entry(self, address):
        return self.entries.get(address)

    def get_container(self, address):
        instructions = self.instructions.get(address)
        if instructions:
            return instructions[0].container

    def get_entry_point_refs(self, address):
        return [referrer.address for referrer in self.instructions[address][0].referrers]

    def _parse_skool(self, skoolfile):
        map_entry = None
        instruction = None
        address_comments = []
        for line in skoolfile:
            if line[0] == ';':
                if self._parse_comment(line):
                    instruction = None
                continue

            if not self.mode.include:
                continue

            if self.debug:
                sys.stderr.write(line)

            s_line = line.strip()
            if not s_line:
                instruction = None
                if self.comments:
                    if map_entry:
                        self._add_end_comment(map_entry)
                    else:
                        self.header += self.comments
                self.comments[:] = []
                self.ignores[:] = []
                map_entry = None
                continue

            if s_line[0] == ';' and map_entry and instruction:
                # This is an instruction comment continuation line
                address_comments[-1][1] = '%s %s' % (address_comments[-1][1], s_line[1:].lstrip())
                continue

            # This line contains an instruction
            instruction, address_comment = self._parse_instruction(line)
            address = instruction.address
            addr_str = instruction.addr_str
            ctl = instruction.ctl
            if ctl in DIRECTIVES:
                desc, details, registers = self._parse_comment_block()
                if ctl == 'c':
                    map_entry = Routine(address, addr_str, ctl, desc, details, registers)
                else:
                    map_entry = Data(address, addr_str, ctl, desc, details)
                map_entry.ignoretua = self.mode.ignoretua
                self.mode.ignoretua = False
                map_entry.ignoredua = self.mode.ignoredua
                self.mode.ignoredua = False
                self.entries[address] = map_entry
                self.memory_map.append(map_entry)
                self.comments[:] = []
            elif ctl == 'd':
                # This is a data definition entry
                map_entry = None
            elif ctl == 'r':
                # This is a remote entry
                map_entry = RemoteEntry(instruction.operation, address)

            if map_entry:
                address_comments.append([instruction, address_comment])
                instructions = self.instructions.setdefault(address, [])
                instructions.append(instruction)
                map_entry.add_instruction(instruction)
                if self.comments:
                    mid_routine_comment = self._join_comments(self.comments, True)
                    map_entry.add_mid_routine_comment(instruction.label, mid_routine_comment)
                    self.comments[:] = []
                    self.mode.ignoremrcua = 0 in self.ignores

            self.mode.apply_asm_attributes(instruction)
            self.ignores[:] = []

            # Set bytes in the snapshot if the instruction is DEF{B,M,S,W}
            if self.mode.html:
                operation = instruction.operation
                if operation[:4].upper() in ('DEFB', 'DEFM', 'DEFS', 'DEFW'):
                    set_bytes(self.snapshot, address, operation)
        if self.comments and map_entry:
            self._add_end_comment(map_entry)

        # Do some post-processing
        self._parse_address_comments(address_comments)
        if self.mode.html:
            self._calculate_entry_sizes()
            self._calculate_references()
            self._escape_instructions()
        else:
            self._substitute_labels()

    def _add_end_comment(self, map_entry):
        map_entry.end_comment = self._join_comments(self.comments, True)
        map_entry.ignoreecua = len(self.ignores) > 0

    def _join_comments(self, comments, split=False):
        sections = [[]]
        for line in comments:
            s_line = line.strip()
            if split and s_line == '.':
                sections.append([])
            elif s_line:
                sections[-1].append(s_line)
        paragraphs = [" ".join(line for line in section) for section in sections if section]
        if self.mode.html:
            paragraphs = [cgi.escape(p) for p in paragraphs]
        if split:
            return paragraphs
        return paragraphs[0] if paragraphs else ''

    def _parse_comment(self, line):
        """Parse a comment line. Return False if the line contains an ASM (@)
        directive, and True otherwise."""
        comment = line[2:].rstrip()
        if comment.startswith('@'):
            self._parse_asm_directive(comment[1:])
            return False
        if self.mode.started and self.mode.include:
            if self.debug:
                sys.stderr.write(line)
            self.comments.append(comment)
            self.mode.ignoreua = False
        return True

    def _parse_asm_directive(self, directive):
        if self.mode.started:
            if directive.startswith('end'):
                self.mode.end()
                return

            prefix = directive[:4]
            infix = directive[len(prefix):len(prefix) + 1]
            suffix = directive[len(prefix) + len(infix):].rstrip()
            if prefix in ('ofix', 'bfix', 'rfix', 'isub', 'rsub') and infix in '+-' and suffix in ('begin', 'else', 'end'):
                cur_op = self.stack[-1] if self.stack else (None, None)
                if suffix == 'begin':
                    if prefix == cur_op[0]:
                        _error('%s inside %s%s block' % (directive, cur_op[0], cur_op[1]))
                    self.stack.append((prefix, infix))
                elif suffix == 'else':
                    if cur_op[0] is None:
                        _error('%s not inside block' % directive)
                    if prefix != cur_op[0] or infix == cur_op[1]:
                        _error('%s inside %s%s block' % (directive, cur_op[0], cur_op[1]))
                    self.stack[-1] = (prefix, infix)
                elif suffix == 'end':
                    if cur_op[0] is None:
                        _error('%s has no matching start directive' % directive)
                    if (prefix, infix) != cur_op:
                        _error('%s cannot end %s block' % (directive, cur_op))
                    self.stack.pop()

                include = True
                for p, i in self.stack:
                    if p == 'ofix':
                        do_op = self.mode.do_ofixes
                    elif p == 'bfix':
                        do_op = self.mode.do_bfixes
                    elif p == 'rfix':
                        do_op = self.mode.do_rfixes
                    elif p == 'isub':
                        do_op = self.mode.asm
                    elif p == 'rsub':
                        do_op = self.mode.do_rsubs
                    include = i == '+' if do_op else i == '-'
                    if not include:
                        break

                self.mode.include = include
                return

            if self.mode.asm and self.mode.include:
                if directive.startswith('rsub='):
                    self.mode.rsub = directive[5:].rstrip()
                elif directive.startswith('ssub='):
                    self.mode.ssub = directive[5:].rstrip()
                elif directive.startswith('isub='):
                    self.mode.isub = directive[5:].rstrip()
                elif directive.startswith('bfix='):
                    self.mode.bfix = directive[5:].rstrip()
                elif directive.startswith('ofix='):
                    self.mode.ofix = directive[5:].rstrip()
                elif directive.startswith('label='):
                    self.mode.label = directive[6:].rstrip()
                elif directive.startswith('nolabel'):
                    self.mode.nolabel = True
                elif directive.startswith('nowarn'):
                    self.mode.nowarn = True
                elif directive.startswith('ignoreua'):
                    self.mode.ignoreua = True
                    self.ignores.append(len(self.comments))
                elif directive.startswith('keep'):
                    self.mode.keep = True
                elif directive.startswith('org='):
                    self.mode.org = directive[4:].rstrip()
        elif directive.startswith('start'):
            self.mode.start()

    def _parse_instruction(self, line):
        ctl = line[0]
        addr_str = line[1:6]
        address = parse_int(addr_str)
        comment_index = line.find(';')
        if comment_index < 0:
            comment_index = len(line)
        operation = line[7:comment_index].strip()
        addr_str, operation = self.mode.apply_case(addr_str, operation)
        addr_str, operation = self.mode.apply_base(addr_str, operation)
        address_comment = line[comment_index + 1:].strip()
        instruction = Instruction(ctl + addr_str, address, operation)
        if instruction.sub:
            self._info('%s %s -> %s' % (address, operation, instruction.sub))
        return instruction, address_comment

    def _apply_ignores(self, section_ignores, index, line_no):
        for i in self.ignores:
            if i < line_no:
                self.ignores.remove(i)
                section_ignores[index] = True
                return

    def _parse_comment_block(self):
        sections = [[], [], []]
        section_ignores = [False, False, False]
        line_no = 0
        index = 0
        last_line = ""
        for line in self.comments:
            if len(line) == 0:
                if len(last_line) > 0:
                    self._apply_ignores(section_ignores, index, line_no)
                    index += 1
            else:
                sections[index].append(line)
                last_line = line
            line_no += 1
        if index < 3:
            self._apply_ignores(section_ignores, index, line_no)
        self.mode.ignoretua = section_ignores[0]
        self.mode.ignoredua = section_ignores[1]
        registers = [reg_line.split(None, 1) for reg_line in sections[2]]
        if self.mode.lower:
            registers = [(n.lower(), d) for n, d in registers]
        elif self.mode.upper:
            registers = [(n.upper(), d) for n, d in registers]
        return [self._join_comments(sections[0]), self._join_comments(sections[1], True), registers]

    def _parse_address_comments(self, comments):
        i = 0
        while i < len(comments):
            instruction, comment = comments[i]
            comment_lines = []
            if comment and comment.strip()[0] == '{':
                nesting = comment.count('{') - comment.count('}')
                while nesting > 0:
                    comment_lines.append(comments[i][1])
                    i += 1
                    nesting += comments[i][1].count('{') - comments[i][1].count('}')
                comment_lines.append(comments[i][1].strip('}'))
                comment_lines[0] = comment_lines[0].strip('{')
            else:
                comment_lines.append(comment)
            rowspan = len(comment_lines)
            address_comment = self._join_comments(comment_lines).strip()
            instruction.set_comment(rowspan, address_comment)
            i += 1

    def _calculate_entry_sizes(self):
        for i, entry in enumerate(self.memory_map[:-1]):
            entry.size = int(self.memory_map[i + 1].address) - int(entry.address)
        last_entry = self.memory_map[-1]
        last_entry.size = 65536 - int(last_entry.address)

    def _calculate_references(self):
        # Parse operations for routine/data addresses
        for entry in self.memory_map:
            for instruction in entry.instructions:
                operation = instruction.operation
                if operation.startswith('DEFW') or operation[0:2] in ['JP', 'JR', 'CA']:
                    addr_str = get_address(operation)
                    if not addr_str:
                        continue
                    address = parse_int(addr_str)
                    other_instructions = self.instructions.get(address, [])
                    if other_instructions:
                        other_entry = other_instructions[0].container
                    elif address in self.entries:
                        other_entry = self.entries[address]
                    else:
                        continue
                    if entry != other_entry and (other_entry.is_remote() or operation.startswith('DEFW') or other_entry.is_routine()):
                        instruction.set_reference(other_entry, address, addr_str)
                        for other_instruction in other_instructions:
                            other_instruction.add_referrer(entry)

    def _escape_instructions(self):
        for entry in self.memory_map:
            for instruction in entry.instructions:
                instruction.html_escape()

    def _info(self, s):
        if self.mode.verbose:
            sys.stderr.write('%s\n' % s)

    def _warn(self, s):
        if self.mode.warn:
            sys.stderr.write('WARNING: %s\n' % s)

    def _substitute_labels(self):
        self._generate_labels()
        for entry in self.memory_map:
            for instruction in entry.instructions:
                if not instruction.sub and instruction.keep:
                    self._info('Keeping operand: %s %s' % (instruction.address, instruction.operation))
                else:
                    self._label_operand(instruction)

    def _generate_labels(self):
        """Generate labels for mid-routine entry points (based on the label of
           the main entry point)."""
        for entry in self.entries.values():
            instructions = entry.instructions
            if instructions:
                main_label = instructions[0].asm_label
                if not main_label and self.mode.create_labels:
                    main_label = instructions[0].asm_label = 'L%s' % entry.addr_str
                if main_label:
                    index = 0
                    for instruction in instructions[1:]:
                        if instruction.ctl in '!*' and not instruction.asm_label and not instruction.nolabel:
                            instruction.asm_label = '%s_%i' % (main_label, index)
                            index += 1

    def _label_operand(self, instruction):
        label_warn = instruction.sub is None and instruction.warn
        operation = instruction.operation
        operand = get_address(operation)
        operand_int = parse_int(operand)
        instructions = None if operand_int is None else self.instructions.get(operand_int)
        if instructions:
            reference = instructions[0]
            if reference.asm_label:
                rep = operation.replace(operand, reference.asm_label)
                if reference.is_in_routine() and label_warn and operation.startswith('LD '):
                    # Warn if a LD operand is replaced with a routine label in
                    # an unsubbed operation (will need @keep to retain operand,
                    # or @nowarn if the replacement is OK)
                    self._warn('LD operand replaced with routine label in unsubbed operation:\n  %s %s -> %s' % (instruction.address, operation, rep))
                instruction.operation = rep
            elif instruction.warn and instruction.is_in_routine():
                # Warn if we cannot find a label to replace the operand of this
                # routine instruction (will need @nowarn if this is OK)
                self._warn('Found no label for operand: %s %s' % (instruction.address, operation))
        elif label_warn and self.mode.do_ssubs and operand and operand_int > 23755:
            # Warn if the operand is > 23755 (where code might be) but doesn't
            # refer to the address of an instruction (will need @nowarn if this
            # is OK)
            self._warn('Unreplaced operand: %s %s' % (instruction.address, operation))

class Mode:
    def __init__(self, case, base, asm_mode, verbose, warn, fix_mode, html, create_labels):
        self.lower = case == CASE_LOWER
        self.upper = case == CASE_UPPER
        self.decimal = base == BASE_10
        self.hexadecimal = base == BASE_16
        self.html = html
        self.asm = asm_mode > 0
        self.verbose = verbose
        self.warn = warn
        self.started = asm_mode == 0
        self.include = self.started
        self.do_rfixes = fix_mode >= 3
        self.do_rsubs = asm_mode >= 3 or self.do_rfixes
        self.do_ssubs = asm_mode >= 2
        self.do_ofixes = fix_mode >= 1 or self.do_rsubs
        self.do_bfixes = fix_mode >= 2
        self.labels = []
        self.create_labels = create_labels
        self.ignoretua = False
        self.ignoredua = False
        self.reset()

    def reset(self):
        self.label = None
        self.isub = None
        self.ofix = None
        self.bfix = None
        self.ssub = None
        self.rsub = None
        self.keep = False
        self.nowarn = False
        self.nolabel = False
        self.ignoreua = False
        self.ignoremrcua = False
        self.org = None

    def start(self):
        self.started = True
        self.include = True

    def end(self):
        if self.asm:
            self.started = False
            self.include = False

    def apply_asm_attributes(self, instruction):
        if self.html:
            return

        if self.label:
            if self.label in self.labels:
                _error('Duplicate label %s at %s' % (self.label, instruction.address))
            self.labels.append(self.label)
            instruction.asm_label = self.label

        sub = self.isub
        if self.bfix is not None and self.do_bfixes:
            sub = self.bfix
        elif self.ofix is not None and self.do_ofixes:
            sub = self.ofix
        elif self.rsub is not None and self.do_rsubs:
            sub = self.rsub
        elif self.ssub is not None and self.do_ssubs:
            sub = self.ssub
        if sub is not None:
            _, sub = self.apply_case('', sub)
            _, sub = self.apply_base('', sub)
            instruction.apply_sub(sub)

        instruction.warn = not self.nowarn
        instruction.nolabel = self.nolabel
        instruction.ignoreua = self.ignoreua
        instruction.ignoremrcua = self.ignoremrcua
        instruction.keep = self.keep
        instruction.org = self.org

        self.reset()

    def apply_case(self, addr_str, operation):
        if self.lower:
            addr_str = addr_str.lower()
            if operation.lower().startswith('defm'):
                operation = 'defm%s' % operation[4:]
            else:
                operation = operation.lower()
        elif self.upper:
            addr_str = addr_str.upper()
            if operation.lower().startswith('defm'):
                operation = 'DEFM%s' % operation[4:]
            else:
                operation = operation.upper()
        return addr_str, operation

    def apply_base(self, addr_str, operation):
        address = parse_int(addr_str)
        if self.decimal:
            if address:
                addr_str = str(address)
            operation = self.convert(operation)
        elif self.hexadecimal:
            if address:
                fmt = '$%04x' if self.lower else '$%04X'
                addr_str = fmt % address
            operation = self.convert(operation)
        return addr_str, operation

    def convert(self, operation):
        if not operation:
            return operation

        ops = operation.upper().split(None, 1)
        if len(ops) > 1:
            ops[1:] = ops[1].split(',')
        ops = [op.strip() for op in ops]

        index = byte = None
        if len(ops) == 2:
            # RL/RR/RRC/RLC/SLA/SRA/SLL/SRL/INC/DEC/AND/OR/XOR/SUB/CP (I[XY]+d)
            index = self.get_index(ops[1])
        elif len(ops) == 3:
            index = self.get_index(ops[1])
            if index:
                # LD (I[XY]+d),n; LD (I[XY]+d),r
                byte = parse_int(ops[2])
            else:
                # LD r,(I[XY]+d); BIT/SET/RES n,(I[XY]+d);
                # ADD/ADC/SBC A,(I[XY]+d)
                index = self.get_index(ops[2])
        if index:
            return self.replace_index(operation, index, byte)

        if ops[0] in ('DEFB', 'DEFS'):
            directive = operation.split(None, 1)[0]
            bytes = [self.replace_byte(b, False) for b in ops[1:]]
            return '%s %s' % (directive, ','.join(bytes))

        if ops[0] == 'DEFW':
            directive = operation.split(None, 1)[0]
            words = [self.replace_address(w, False) for w in ops[1:]]
            return '%s %s' % (directive, ','.join(words))

        if ops[0] in ('CALL', 'DJNZ', 'JP', 'JR'):
            # CALL [*,]nn, DJNZ nn, JP [*,]nn, JR [*,]nn
            return self.replace_address(operation)

        if ops[0] in ('AND', 'OR', 'XOR', 'SUB', 'CP', 'IN', 'OUT', 'ADD', 'ADC', 'SBC', 'RST'):
            # AND n; OR n; XOR n; SUB n; CP n; IN (n),A; OUT (n),A; ADD A,n;
            # ADC A,n; SBC A,n; RST n
            return self.replace_byte(operation)

        if ops[0] == 'LD':
            if ops[1] in ('A', 'B', 'C', 'D', 'E', 'H', 'L', 'IXL', 'IXH', 'IYL', 'IYH', '(HL)') and not ops[2].startswith('('):
                # LD r,n; LD (HL),n
                return self.replace_byte(operation)
            addr_ld_reg = ('A', 'BC', 'DE', 'HL', 'IX', 'IY', 'SP')
            if ops[1] in addr_ld_reg or ops[2] in addr_ld_reg:
                # LD A,(nn); LD (nn),A; LD rr,nn; LD rr,(nn); LD (nn),rr
                return self.replace_address(operation)

        return operation

    def get_index(self, op):
        return re.search('\(I[XY] *[\+-].*\)', op)

    def replace_index(self, operation, index, byte):
        match = index.group()
        marker = '_'
        operation = operation.replace(match, marker)
        if byte is not None:
            operation = self.replace_byte(operation)
        return operation.replace(marker, self.replace_byte(match))

    def replace_byte(self, text, check_prefix=True):
        return self.replace_number(text, '$%02X', check_prefix)

    def replace_address(self, text, check_prefix=True):
        return self.replace_number(text, '$%04X', check_prefix)

    def replace_number(self, text, hex_fmt, check_prefix):
        num_str = get_address(text, check_prefix)
        num = parse_int(num_str)
        if num is not None:
            if self.decimal:
                return text.replace(num_str, str(num))
            elif self.hexadecimal:
                if self.lower:
                    hex_fmt = hex_fmt.lower()
                return text.replace(num_str, hex_fmt % num)
        return text

class Instruction:
    def __init__(self, label, address, operation):
        self.label = label
        self.ctl = label[0]
        self.addr_str = label[1:] if label[1].isdigit() else label[2:]
        self.address = address
        self.operation = operation
        self.container = None
        self.reference = None
        self.comment = None
        self.referrers = []
        self.asm_label = None
        # If this instruction has no address, it was inserted between
        # @rsub+begin and @rsub+end; in that case, mark it as a subbed
        # instruction already
        self.sub = None if address else operation
        self.keep = False
        self.warn = True
        self.ignoreua = False
        self.ignoremrcua = False

    def set_comment(self, rowspan, text):
        self.comment = Comment(rowspan, text)

    def set_reference(self, entry, address, addr_str):
        self.reference = Reference(entry, address, addr_str)

    def add_referrer(self, routine):
        if routine not in self.referrers:
            self.referrers.append(routine)
        self.container.add_referrer(routine)

    def apply_sub(self, sub):
        self.sub = sub
        self.operation = sub

    def is_in_routine(self):
        return self.container.is_routine()

    def get_mid_routine_comment(self):
        return self.container.get_mid_routine_comment(self.label)

    def html_escape(self):
        self.operation = cgi.escape(self.operation)

class Reference:
    def __init__(self, entry, address, addr_str):
        self.entry = entry
        self.address = address
        self.addr_str = addr_str

class Comment:
    def __init__(self, rowspan, text):
        self.rowspan = rowspan
        self.text = text

class SkoolEntry:
    def __init__(self, address, addr_str=None, ctl=None, description=None, details=()):
        self.asm_id = None
        self.address = address
        self.addr_str = addr_str
        self.ctl = ctl
        self.description = description
        self.details = details
        self.registers = []
        self.instructions = []
        self.mid_routine_comments = {}
        self.end_comment = ()
        self.referrers = []
        self.size = None
        self.ignoretua = False
        self.ignoredua = False
        self.ignoreecua = False

    def is_routine(self):
        return False

    def is_data(self):
        return False

    def is_remote(self):
        return False

    def add_instruction(self, instruction):
        instruction.container = self
        self.instructions.append(instruction)

    def get_mid_routine_comment(self, address):
        return self.mid_routine_comments.get(address, ())

    def add_mid_routine_comment(self, address, text):
        self.mid_routine_comments[address] = text

    def add_referrer(self, routine):
        if routine not in self.referrers:
            self.referrers.append(routine)

class RemoteEntry(SkoolEntry):
    def __init__(self, asm_id, address):
        SkoolEntry.__init__(self, address)
        self.asm_id = asm_id

    def is_remote(self):
        return True

class Routine(SkoolEntry):
    def __init__(self, address, addr_str, ctl=None, description=None, details=None, registers=()):
        SkoolEntry.__init__(self, address, addr_str, ctl, description, details)
        for name, contents in registers:
            self.registers.append(Register(name, contents))

    def is_routine(self):
        return True

class Register:
    def __init__(self, name, contents):
        self.name = name
        self.contents = contents

class Data(SkoolEntry):
    def is_data(self):
        return True

class TableParser:
    TABLE_MARKER = '#TABLE'
    TABLE_END_MARKER = 'TABLE#'
    COLUMN_WRAP_MARKER = ':w'

    def parse_text(self, text, index):
        end_marker = self.__class__.TABLE_END_MARKER
        end = text.index(end_marker, index) + len(end_marker)
        table = self.parse_table(text[index:end])
        return end, table

    def parse_table(self, table_def):
        text = table_def
        for ws_char in '\n\r\t':
            text = text.replace(ws_char, ' ')

        marker = self.__class__.TABLE_MARKER
        index = len(marker) if text.startswith(marker) else 0
        classes = []
        if text[index] == '(':
            end = text.find(')', index)
            if end < 0:
                _error("Cannot find closing ')' in table CSS class list:\n%s\n" % table_def)
            classes = [c.strip() for c in text[index + 1:end].split(',')]
            index = end + 1
        table_class = classes[0] if classes else ''
        column_classes = classes[1:]
        wrap_columns = []
        for i, column_class in enumerate(column_classes):
            if column_class.endswith(self.__class__.COLUMN_WRAP_MARKER):
                column_classes[i] = column_class[:-len(self.__class__.COLUMN_WRAP_MARKER)]
                wrap_columns.append(i)
        table = Table(table_class, wrap_columns)

        prev_spans = {}
        while text.find('{', index) >= 0:
            row = []
            row_start = text.find('{ ', index)
            if row_start < 0:
                _error("Cannot find opening '{ ' in table row:\n%s\n" % table_def)
            row_end = text.find(' }', row_start)
            if row_end < 0:
                _error("Cannot find closing ' }' in table row:\n%s\n" % table_def)
            col_index = 0
            for cell in text[row_start + 1:row_end + 1].split(' | '):
                prev_rowspan, prev_colspan = prev_spans.get(col_index, (1, 1))
                while prev_rowspan > 1:
                    prev_spans[col_index] = (prev_rowspan - 1, prev_colspan)
                    col_index += prev_colspan
                    prev_rowspan, prev_colspan = prev_spans.get(col_index, (1, 1))
                header, transparent = False, False
                rowspan, colspan = 1, 1
                cell_class = column_classes[col_index] if len(column_classes) > col_index else ''
                cell = cell.strip()
                if cell.startswith('='):
                    end = cell.find(' ')
                    if end < 0:
                        end = len(cell)
                    for span in cell[1:end].split(','):
                        if span[0] == 'c':
                            colspan = int(span[1:])
                        elif span[0] == 'r':
                            rowspan = int(span[1:])
                        elif span[0] == 'h':
                            header = True
                        elif span[0] == 't':
                            transparent = True
                    cell = cell[end:].lstrip()
                row.append(Cell(cell, transparent, colspan, rowspan, header, cell_class))
                prev_spans[col_index] = (rowspan, colspan)
                col_index += colspan

            # Deal with the case where the previous row contains one or more
            # cells at the end with rowspan > 1
            while col_index in prev_spans:
               prev_rowspan, prev_colspan = prev_spans[col_index]
               if prev_rowspan > 1:
                   prev_spans[col_index] = (prev_rowspan - 1, prev_colspan)
               col_index += prev_colspan

            table.add_row(row)
            index = row_end + 1

        return table

class Table:
    def __init__(self, table_class, wrap_columns):
        self.table_class = table_class
        self.wrap_columns = wrap_columns
        self.rows = []
        self.cells = []
        self.col_widths = None
        self.cell_padding = None
        self.num_cols = None

    def add_row(self, cells):
        self.rows.append(cells)
        self.cells.extend(cells)

    def prepare_cells(self):
        # For each cell, set row_index and col_index, and convert the contents
        # to a 1-item list. Also calculate num_cols and col_widths.
        prev_spans = {}
        for row_index, row in enumerate(self.rows):
            col_index = 0
            for cell in row:
                while True:
                    prev_rowspan, prev_colspan = prev_spans.get(col_index, (1, 1))
                    if prev_rowspan == 1:
                        break
                    prev_spans[col_index] = (prev_rowspan - 1, prev_colspan)
                    col_index += prev_colspan
                cell.row_index, cell.col_index = row_index, col_index
                cell.contents = [cell.contents]
                prev_spans[col_index] = (cell.rowspan, cell.colspan)
                col_index += cell.colspan

            # Deal with cells at the end of the previous row that have rowspan
            # greater than 1
            while col_index in prev_spans:
               prev_rowspan, prev_colspan = prev_spans[col_index]
               if prev_rowspan > 1:
                   prev_spans[col_index] = (prev_rowspan - 1, prev_colspan)
               col_index += prev_colspan

        self.num_cols = 1 + max([cell.col_index for cell in [row[-1] for row in self.rows]])
        self._calculate_col_widths()

    def get_header_rows(self):
        headers = []
        for i, row in enumerate(self.rows):
            if len([c for c in row if not (c.transparent or c.header)]) == 0:
                headers.append(i)
        return headers

    def get_cell_width(self, col_index, colspan):
        return sum(self.col_widths[col_index:col_index + colspan]) + self.cell_padding * (colspan - 1)

    def _calculate_col_widths(self):
        # Loop over the cells to collect their widths (considering only those
        # cells with colspan=1 on this pass)
        cell_widths = [[0] * self.num_cols for row in self.rows]
        for cell in [c for c in self.cells if c.colspan == 1]:
            cell_widths[cell.row_index][cell.col_index] = cell.get_width()

        # Scan each column to find the widest cell; the width of that cell will
        # be the width of the column
        self.col_widths = [max([row[i] for row in cell_widths]) for i in range(self.num_cols)]

        # Scan the cells again to make sure that those with colspan > 1 have
        # enough space
        for cell in [c for c in self.cells if c.colspan > 1]:
            col_index, colspan = cell.col_index, cell.colspan
            space_needed = cell.get_width() - self.get_cell_width(col_index, colspan)
            while space_needed > 0:
                for j in range(colspan):
                    if space_needed > 0:
                        self.col_widths[col_index + j] += 1
                        space_needed -= 1

    def reduce_width(self, max_table_width, min_col_width):
        # Reduce the width of the wrappable columns until either the table is
        # narrow enough or all the wrappable columns are the minimum width
        done = False
        while self.get_width() > max_table_width and not done:
            done = True
            for wrap_col in self.wrap_columns:
                if self.col_widths[wrap_col] > min_col_width:
                    self.col_widths[wrap_col] -= 1
                    done = False

        # Wrap the contents of the cells in the wrappable columns
        for cell in [c for c in self.cells if c.col_index in self.wrap_columns]:
            width = self.get_cell_width(cell.col_index, cell.colspan)
            cell.contents = textwrap.wrap(cell.contents[0], width) or ['']

        # The column widths need to be recalculated now
        self._calculate_col_widths()

    def get_width(self):
        padding = self.cell_padding * (self.num_cols + 1) - 2
        return padding + sum(self.col_widths)

class Cell:
    def __init__(self, contents, transparent, colspan, rowspan, header, cell_class):
        self.contents = contents
        self.colspan = colspan
        self.rowspan = rowspan
        self.header = header
        self.transparent = transparent
        self.cell_class = cell_class
        self.row_index = None
        self.col_index = None

    def get_width(self):
        return max([len(line) for line in self.contents])

class RefParser:
    def __init__(self):
        self.ref = []
        self.sections = {}

    def parse(self, reffile):
        infile = open(reffile, 'r')
        section_name = None
        section_lines = []
        for line in infile:
            if line.startswith('['):
                if section_name:
                    self.ref.append((section_name, section_lines))
                    self.sections[section_name] = section_lines
                section_name = line[1:line.index(']')]
                section_lines = []
            elif not line.startswith(';'):
                section_lines.append(line.rstrip())
        infile.close()
        if section_name:
            self.ref.append((section_name, section_lines))
            self.sections[section_name] = section_lines

    def has_ref_items(self, section_type):
        prefix = section_type + ':'
        for section_name in self.sections:
            if section_name.startswith(prefix):
                return True
        return False

    def build_dictionary(self, section_name):
        return self._get_dictionary(self.sections.get(section_name, ()))

    def build_dictionaries(self, section_type):
        sep = ':'
        prefix = section_type + sep
        dictionaries = []
        for section_name, lines in self.ref:
            if section_name.startswith(prefix):
                section_id = section_name.split(sep, 1)[1]
                dictionaries.append((section_id, self._get_dictionary(lines)))
        return dictionaries

    def get_section(self, section_name):
        paragraphs = self._get_paragraphs(self.sections.get(section_name, ()), True)
        lines = ['\n'.join(p) for p in paragraphs]
        return '\n'.join(lines)

    def get_data_dictionary(self, data_name):
        return self.build_dictionary('Data:%s' % data_name)

    def get_ref_items(self, section_type, keep_lines=False, as_list=False):
        sep = ':'
        prefix = section_type + sep
        items = []
        for section_name, lines in self.ref:
            if section_name.startswith(prefix):
                if as_list:
                    contents = self._get_list(lines)
                else:
                    contents = self._get_paragraphs(lines, keep_lines)
                elements = section_name.split(sep, 2)
                if len(elements) < 3:
                    items.append((elements[1], contents))
                else:
                    items.append((elements[1], elements[2], contents))
        return items

    def get_list(self, section_name):
        return self._get_list(self.sections.get(section_name, ()))

    def _get_list(self, lines):
        items = []
        for line in lines:
            sline = line.lstrip()
            if sline:
                items.append(sline)
        return items

    def _get_dictionary(self, lines):
        dictionary = {}
        for line in lines:
            if line:
                key, value = line.split('=', 1)
                if key.isdigit():
                    key = int(key)
                dictionary[key] = value
        return dictionary

    def _get_paragraphs(self, lines, keep_lines):
        paragraphs = []
        p = []
        for line in lines:
            if line:
                p.append(line)
            elif p:
                paragraphs.append(p if keep_lines else '\n'.join(p))
                p = []
        if p:
            paragraphs.append(p if keep_lines else '\n'.join(p))
        return paragraphs
